from reling.data import LANGUAGES_PATH, SPEAKERS_PATH, STYLES_PATH, TOPICS_PATH
from reling.db.helpers.languages import populate_languages
from reling.db.helpers.modifiers import populate_modifiers
from reling.db.models import Speaker, Style, Topic
from reling.utils.paths import get_app_data_parent
from . import commands  # Register commands with Typer
from .app import app
from .db import init_db

__all__ = [
    'app',
]

APP_NAME = 'ReLing'
DB_NAME = 'reling.db'

# This code must run both during app execution and on auto-completion.
# Therefore, it should be placed at the top level of the module.
DATA_PATH = get_app_data_parent() / APP_NAME
DATA_PATH.mkdir(parents=True, exist_ok=True)
init_db(DATA_PATH / DB_NAME)
populate_languages(LANGUAGES_PATH)
populate_modifiers(Topic, TOPICS_PATH)
populate_modifiers(Style, STYLES_PATH)
populate_modifiers(Speaker, SPEAKERS_PATH)



from reling.app.app import app
from reling.app.translation import get_dialogue_exchanges, get_text_sentences
from reling.app.types import (
    API_KEY,
    CONTENT_ARG,
    LANGUAGE_OPT_ARG,
    MODEL,
    READ_OPT,
    TTS_MODEL,
)
from reling.db.models import Dialogue, Language, Text
from reling.gpt import GPTClient
from reling.helpers.output import output, SentenceData
from reling.helpers.voices import pick_voice, pick_voices
from reling.tts import TTSClient

__all__ = [
    'show',
]

SPEAKER_PREFIX = '> '
USER_PREFIX = '< '


@app.command()
def show(
        api_key: API_KEY,
        model: MODEL,
        tts_model: TTS_MODEL,
        content: CONTENT_ARG,
        language: LANGUAGE_OPT_ARG = None,
        read: READ_OPT = False,
) -> None:
    """Display a text or dialogue, or its translation if a language is specified."""
    (show_text if isinstance(content, Text) else show_dialogue)(
        GPTClient(api_key=api_key, model=model),
        content,
        language or content.language,
        TTSClient(api_key=api_key, model=tts_model.get()) if read else None,
    )


def show_text(gpt: GPTClient, text: Text, language: Language, tts: TTSClient | None) -> None:
    """Display the text in the specified language, optionally reading it out loud."""
    voice_tts = tts.with_voice(pick_voice()) if tts else None
    for sentence in get_text_sentences(gpt, text, language):
        output(SentenceData.from_tts(sentence, voice_tts))


def show_dialogue(gpt: GPTClient, dialogue: Dialogue, language: Language, tts: TTSClient | None) -> None:
    """Display the dialogue in the specified language, optionally reading it out loud."""
    exchanges = get_dialogue_exchanges(gpt, dialogue, language)
    speaker_tts, user_tts = map(tts.with_voice, pick_voices(
        dialogue.speaker_gender,
        dialogue.user_gender,
    )) if tts else (None, None)
    for exchange in exchanges:
        output(SentenceData.from_tts(exchange.speaker, speaker_tts, print_prefix=SPEAKER_PREFIX))
        output(SentenceData.from_tts(exchange.user, user_tts, print_prefix=USER_PREFIX))



__all__ = [
    'replace_prefix_casing',
]


def replace_prefix_casing(string: str, prefix: str) -> str:
    """Replace the prefix of a string with the casing of the given prefix."""
    return prefix + string[len(prefix):] if string.lower().startswith(prefix.lower()) else string



from pathlib import Path

from openai import OpenAI

from reling.db.models import Language
from reling.types import Transcriber
from reling.utils.openai import openai_handler

__all__ = [
    'ASRClient',
]


class ASRClient:
    _client: OpenAI
    _model: str

    def __init__(self, *, api_key: str, model: str) -> None:
        self._client = OpenAI(api_key=api_key)
        self._model = model

    def transcribe(self, file: Path, language: Language | None = None, context: str | None = None) -> str:
        """Transcribe an audio file."""
        with openai_handler():
            return self._client.audio.transcriptions.create(
                file=file.open('rb'),
                model=self._model,
                language=language.short_code if language else None,
                prompt=context,
            ).text

    def get_transcriber(self, language: Language | None = None, context: str | None = None) -> Transcriber:
        def transcribe(file: Path) -> str:
            return self.transcribe(file, language, context)
        return transcribe



from typing import Iterable

from rich.console import Console, JustifyMethod
from rich.table import Table

__all__ = [
    'build_table',
    'DEFAULT_JUSTIFY_METHOD',
    'JustifyMethod',
    'Table',
    'print_table',
]

DEFAULT_JUSTIFY_METHOD: JustifyMethod = 'left'


def build_table(
        *,
        title: str | None = None,
        headers: list[str],
        data: Iterable[dict[str, str]],
        justify: dict[str, JustifyMethod] | None = None,
        group_by: list[str] | None = None,
) -> Table:
    """
    Build a Rich Table from headers and data.

    :param title: Table title.
    :param headers: List of column headers.
    :param data: Iterable of dictionaries with data. Keys must match headers.
    :param justify: Dictionary of column justifications. Keys must be in headers. Default is DEFAULT_JUSTIFY_METHOD.
    :param group_by: List of columns to group by: rows with the same values in these columns will be grouped together.
                     Only the first row of each group will have the group values displayed.
    :return: Rich Table.
    """
    justify = justify or {}
    if any(col not in headers for col in justify.keys()):
        raise ValueError('Justify columns must be in headers.')

    table = Table(title=title)
    for header in headers:
        table.add_column(header, justify=justify.get(header, DEFAULT_JUSTIFY_METHOD))

    group_by = group_by or []
    if any(col not in headers for col in group_by):
        raise ValueError('Group by columns must be in headers.')
    last_group: tuple[str, ...] | None = None

    for row in data:
        if sorted(row.keys()) != sorted(headers):
            raise ValueError('Data keys must match headers.')
        group = tuple(row[col] for col in group_by)
        if new_section := group != last_group:
            last_group = group
            table.add_section()
        table.add_row(*[row[col]
                        if new_section or col not in group_by
                        else ''
                        for col in headers])

    return table


def print_table(table: Table) -> None:
    Console().print(table)



from pathlib import Path

from reling.db import init_db as do_init_db

__all__ = [
    'get_db_path',
    'init_db',
]

DB_PATH: Path | None = None


def init_db(path: Path) -> None:
    global DB_PATH
    if DB_PATH is not None:
        raise RuntimeError('Database is already initialized')
    DB_PATH = path
    do_init_db(f'sqlite:///{DB_PATH}')


def get_db_path() -> Path:
    if DB_PATH is None:
        raise RuntimeError('Database is not initialized')
    return DB_PATH



from itertools import islice
from typing import Generator

from reling.app.exceptions import AlgorithmException
from reling.db.enums import ContentCategory
from reling.db.models import Language
from reling.gpt import GPTClient
from reling.types import DialogueExchangeData
from reling.utils.english import pluralize
from reling.utils.iterables import group_items
from reling.utils.transformers import add_numbering, apply, omit_empty, remove_numbering, strip
from .types import ExchangeWithTranslation, ScoreWithSuggestion, SentenceWithTranslation

__all__ = [
    'MAX_SCORE',
    'score_dialogue_translations',
    'score_text_translations',
]

MAX_SCORE = 10
NA = 'N/A'
EMPTY_TRANSLATION = '<empty>'


def build_prompt(
        category: ContentCategory,
        source_language: Language,
        target_language: Language,
        blocks: list[str],
        translations: list[str],
) -> str:
    """Build a prompt for scoring translations."""
    # Speaker turns in dialogues are "graded" as well so that the model appreciates the context.
    n = len(blocks)
    return '\n'.join([
        f'Below {'is' if n == 1 else 'are'} {n} {pluralize('sentence', n)} from a {category.value} '
        f'in {source_language.name} along with {'its' if n == 1 else 'their'} {pluralize('translation', n)} '
        f'into {target_language.name} made by a language learner.',

        f'Score {'the' if n == 1 else 'each'} translation on a scale from 0 to {MAX_SCORE}. '
        f'If {'the' if n == 1 else 'a'} translation is empty, very short, or poor, assign a low score. ',
        f'If the translation is less than perfect, suggest a minimally modified version that would '
        f'deserve a {MAX_SCORE}.',

        f'{'Provide' if n == 1 else 'For each translation, provide'} your feedback on exactly four lines:',
        f'- original sentence on the first line;',  # The first two lines help improve the model's performance
        f'- learner\'s translation on the second line;',
        f'- score (just the number) on the third line;',
        f'- suggested modified translation (or "{NA}") on the fourth line.',

        *([f'Provide this feedback for each of the {n} translations.'] if n > 1 else []),
        f'Say nothing else.',
        f'',
        f'The original {category.value} is:',
        *apply(add_numbering, blocks),
        f'',
        f'The translations are:',
        *apply(add_numbering, [translation.strip() or EMPTY_TRANSLATION for translation in translations]),
    ])


def ask_and_parse(gpt: GPTClient, prompt: str) -> Generator[ScoreWithSuggestion, None, None]:
    """
    Ask the model to score translations and parse the output.
    :raises AlgorithmException: If there is an issue with the output of the model.
    """
    for _, _, string_score, suggestion in group_items(gpt.ask(
        prompt,
        creative=False,
        transformers=[strip, omit_empty, remove_numbering],
    ), 4):
        try:
            score = int(string_score)
        except ValueError:
            raise AlgorithmException(f'Could not parse the score as an integer from the model output: {string_score}.')
        if score < 0 or score > MAX_SCORE:
            raise AlgorithmException(f'The score {score} given by the model is not in the range from 0 to {MAX_SCORE}.')
        yield ScoreWithSuggestion(
            score=score,
            suggestion=suggestion if suggestion != NA and score < MAX_SCORE else None,
        )


def score_text_translations(
        gpt: GPTClient,
        sentences: list[SentenceWithTranslation],
        source_language: Language,
        target_language: Language,
) -> Generator[ScoreWithSuggestion, None, None]:
    """
    Score the translations of a text and provide suggestions for improvement.
    :raises AlgorithmException: If there is an issue with the scoring algorithm.
    """
    prompt = build_prompt(
        category=ContentCategory.TEXT,
        source_language=source_language,
        target_language=target_language,
        blocks=[sentence.sentence for sentence in sentences],
        translations=[sentence.translation.text for sentence in sentences],
    )
    return ask_and_parse(gpt, prompt)


def score_dialogue_translations(
        gpt: GPTClient,
        exchanges: list[ExchangeWithTranslation],
        original_translations: list[DialogueExchangeData],
        source_language: Language,
        target_language: Language,
) -> Generator[ScoreWithSuggestion, None, None]:
    """
    Score the translations of user turns in a dialogue and provide suggestions for improvement.
    :raises AlgorithmException: If there is an issue with the scoring algorithm.
    """
    prompt = build_prompt(
        category=ContentCategory.DIALOGUE,
        source_language=source_language,
        target_language=target_language,
        blocks=[turn for exchange in exchanges for turn in exchange.exchange.all()],
        translations=[turn
                      for exchange, original_translation in zip(exchanges, original_translations)
                      for turn in [original_translation.speaker, exchange.user_translation.text]],
    )
    yield from islice(ask_and_parse(gpt, prompt), 1, None, 2)



from .audio import Reader, Speed, Transcriber
from .dialogue_exchange_data import DialogueExchangeData
from .input import Input
from .word_with_sense import WordWithSense

__all__ = [
    'DialogueExchangeData',
    'Input',
    'Reader',
    'Speed',
    'Transcriber',
    'WordWithSense',
]



from typing import Callable, cast, Generator, Iterable

__all__ = [
    'group_items',
    'map_asterisk',
    'pair_items',
]


def group_items[T](iterable: Iterable[T], group_size: int) -> Generator[tuple[T, ...], None, None]:
    """
    Group items from the iterable into tuples of the specified size.
    The last tuple is discarded if it is shorter than the specified size.
    """
    items = iter(iterable)
    return zip(*(items for _ in range(group_size)))


def pair_items[T](iterable: Iterable[T]) -> Generator[tuple[T, T], None, None]:
    """Pair items from the iterable into tuples. The last item is discarded if there is an odd number of items."""
    return cast(Generator[tuple[T, T], None, None], group_items(iterable, 2))


def map_asterisk[T](f: Callable[..., T], items: Iterable[tuple]) -> Generator[T, None, None]:
    """Apply the function to each tuple in the iterable as positional arguments."""
    return (f(*args) for args in items)



from sqlalchemy import Index
from sqlalchemy.orm import Mapped, mapped_column
from sqlalchemy.sql import func

from reling.db.base import Base

__all__ = [
    'Language',
]


class Language(Base):
    __tablename__ = 'languages'

    id: Mapped[str] = mapped_column(primary_key=True)
    short_code: Mapped[str] = mapped_column(index=True)
    name: Mapped[str]
    extra_name_a: Mapped[str | None]
    extra_name_b: Mapped[str | None]

    __table_args__ = (
        Index('language_name', func.lower('name')),
        Index('language_extra_name_a', func.lower('extra_name_a')),
        Index('language_extra_name_b', func.lower('extra_name_b')),
    )



from reling.db import single_session
from reling.db.models import Dialogue, Text

__all__ = [
    'find_content',
]


def find_content(name: str) -> Text | Dialogue | None:
    """Find a text or dialogue by its name."""
    with single_session() as session:
        return session.get(Text, name) or session.get(Dialogue, name)



from itertools import count

from reling.db import single_session
from reling.db.enums import ContentCategory, Gender, Level
from reling.db.helpers.ids import find_ids_by_prefix
from reling.db.models import Dialogue, DialogueExchange, IdIndex, Language, Text, TextSentence
from reling.types import DialogueExchangeData
from reling.utils.time import now


__all__ = [
    'save_dialogue',
    'save_text',
]


def generate_id(suggested_id: str) -> str:
    """Generate a unique ID based on the suggested ID."""
    taken_ids = set(find_ids_by_prefix(suggested_id))
    if suggested_id not in taken_ids:
        return suggested_id
    for suffix in count(2):
        if (suffixed_id := f'{suggested_id}-{suffix}') not in taken_ids:
            return suffixed_id


def save_text(
        suggested_id: str,
        sentences: list[str],
        language: Language,
        level: Level,
        topic: str,
        style: str,
) -> str:
    """Save a text with the given sentences and return its ID."""
    with single_session() as session:
        text_id = generate_id(suggested_id)
        text = Text(
            id=text_id,
            language_id=language.id,
            level=level,
            topic=topic,
            style=style,
            created_at=now(),
            archived_at=None,
        )
        session.add(text)
        session.add_all([
            TextSentence(
                text_id=text.id,
                index=index,
                sentence=sentence,
            )
            for index, sentence in enumerate(sentences)
        ])
        session.add(IdIndex(
            id=text_id,
            category=ContentCategory.TEXT,
        ))
        session.commit()
    return text_id


def save_dialogue(
        suggested_id: str,
        exchanges: list[DialogueExchangeData],
        language: Language,
        level: Level,
        speaker: str,
        topic: str | None,
        speaker_gender: Gender,
        user_gender: Gender,
) -> str:
    """Save a dialogue with the given exchanges and return its ID."""
    with single_session() as session:
        dialogue_id = generate_id(suggested_id)
        dialogue = Dialogue(
            id=dialogue_id,
            language_id=language.id,
            level=level,
            speaker=speaker,
            topic=topic,
            speaker_gender=speaker_gender,
            user_gender=user_gender,
            created_at=now(),
            archived_at=None,
        )
        session.add(dialogue)
        session.add_all([
            DialogueExchange(
                dialogue_id=dialogue.id,
                index=index,
                speaker=exchange.speaker,
                user=exchange.user,
            )
            for index, exchange in enumerate(exchanges)
        ])
        session.add(IdIndex(
            id=dialogue_id,
            category=ContentCategory.DIALOGUE,
        ))
        session.commit()
    return dialogue_id
